/*
 * Copyright (c) 2003 Lev Walkin <vlm@lionet.info>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <paths.h>
#include <sched.h>
#include <sysexits.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/event.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>
#include <libgen.h>

/*
 * Prepare a kernel queue.
 */
int urun_prepare(int orig_pid, int *mutex_fds);

/*
 * A process that monitors other processes.
 */
void urun_daemon(int kernel_queue);

/*
 * Report a process status.
 */
int proc_status(pid_t pid, int status);

/*
 * Kill all recorded processes.
 * If allow_warn is set, the processes will be notified with SIGTERM.
 */
void kill_recorded_processes(int allow_warn);

/*
 * Storage for the running threads.
 */
typedef struct pidstore_s {
	pid_t pid;			/* Stored PID of the running process */
	int warned;			/* Notified with SIGTERM */
	struct pidstore_s *next;
} pidstore_t;
#define	RP_BUCKETS	(4096)
pidstore_t running_pids[RP_BUCKETS];
pidstore_t *free_psts;	/* Link of empty pid storages */
int running_pids_count;

/*
 * int main.
 */
int
main(int argc, char **argv) {
	char *shell_to_run;
	int pipe_fds[2];
	pid_t original_pid;
	pid_t pid;
	int ret;

	/*
	 * Prepare everything before forking off the monitor.
	 * - remember the original PID;
	 * - create a mutex (permission pipe);
	 */
	original_pid = getpid();
	if(pipe(pipe_fds) == -1) {
		perror("pipe() failed");
		exit(EX_OSERR);
	}

	/*
	 * Fork off a monitor process.
	 */
	pid = fork();
	if(pid == -1) {
		perror("fork() failed");
		exit(EX_OSERR);
	} else if(pid == 0) {
		/*
		 * A child is the monitor process.
		 */
		int q;

		q = urun_prepare(original_pid, pipe_fds);
		/* urun_prepare() won't return if (q == -1) */

		urun_daemon(q);

		_exit(EX_OSERR);	/* The last resort */
	} else {
		/*
		 * A parent.
		 * Down there we'll execute the shell.
		 * Meanwhile, cleanup from forking.
		 */

		/*
		 * Drop privileges.
		 */
		setgid(getgid());
		setuid(getuid());

		/*
		 * Wait for permission to run.
		 * If we are not doing this, we could miss the fork()
		 * in the monitor. So, make a chance to initialize the
		 * monitor before we could fork().
		 */
		close(pipe_fds[1]);	/* Close the write end */
		/* Wait for the permission to go further */
		do {
			ret = read(pipe_fds[0], &ret, 1);
		} while(ret == -1 && errno == EINTR);
		close(pipe_fds[0]);	/* Close the read end */
		if(ret == 0) {
			/*
			 * Something bad happenned to the monitor.
			 * Don't allow to go futher.
			 */
			_exit(EX_OSERR);
		}
	}

	/*
	 * Prepare the shell to run.
	 */

	/*
	 * Change the shell, if requested.
	 */
	shell_to_run = getenv("URUN_PROG");
	if(!shell_to_run) {
		char buf[256];
		ret = readlink("/etc/urun.shell", buf, sizeof(buf));
		if(ret > 1 && ret < sizeof(buf)) {
			/* Readlink does not terminate name. */
			buf[ret] = '\0';
			shell_to_run = buf;
		} else {
			shell_to_run = _PATH_BSHELL;
		}
	}

	/*
	 * If the shell won't change the title, do it myself.
	 */
	setproctitle("-%s (%s)",
		basename(shell_to_run),
		(argc>0) ? basename(argv[0]) : "<URUN>");

	/*
	 * Execute the shell.
	 */
	execl(shell_to_run, "-", NULL);
	perror("execl() failed");

	return (EX_OSERR);
}


/*
 * Prepare the monitor process.
 */
int
urun_prepare(int orig_pid, int *mutex_fds) {
	struct rlimit infinity;
	int q;
	int max_fd;
	int i;

	/*
	 * Record the main process to the list of running processes.
	 */
	if(proc_status(orig_pid, 1))
		_exit(EX_OSERR);

	max_fd = getdtablesize();
	max_fd += 3;	/* Just in case */

	(void)setsid();		/* Turn into a session leader. */

	q = kqueue();
	if(q == -1) {
		_exit(EX_OSERR);
	} else {
		struct kevent kev;
		int ret;

		EV_SET(&kev, orig_pid, EVFILT_PROC,
			EV_ADD | EV_ENABLE,
			NOTE_EXIT
			| NOTE_TRACK
			| NOTE_TRACKERR
			, 0, 0);

		ret = kevent(q, &kev, 1, NULL, 0, NULL);
		if(ret == -1) {
			if(errno == ESRCH) {
				/*
				 * This can happen only if the main process
				 * crashes while waiting on mutex_fd[0].
				 * Just ignore it: no harm made.
				 */
			}
			_exit(EX_OSERR);
		}
	}

	close(mutex_fds[0]);	/* Close read end, just in case */

	/*
	 * Close the file descriptors.
	 */
	for(i = 0; i < max_fd; i++) {
		if(i != mutex_fds[1]
		&& i != q
		&& i != 2	/* Keep it around to notify user */
		) {
			close(i);
		}
	}

	/*
	 * If this user is very restricted,
	 * it is possible to create so many processes
	 * that urun will loose some.
	 * Albeit unlikely, let's not play chances.
	 */
	infinity.rlim_cur = RLIM_INFINITY;
	infinity.rlim_max = RLIM_INFINITY;
	(void)setrlimit(RLIMIT_DATA, &infinity);

	/*
	 * Send the notification that it is now safe to go.
	 */
	write(mutex_fds[1], "Go!", 1);	/* EINTR? Let's it be a bug. */
	close(mutex_fds[1]);	/* Close the write end */

	return q;
}

/*
 * A process that monitors other processes.
 */
void
urun_daemon(int q) {
	const time_t normal_timeout = 5;
	struct kevent kevl[128];
	struct timespec tmo;
	time_t max_allowed_time;
	time_t urun_timeout;
	int war_mode = 0;	/* Extermination of everything. */
	int ret;
	char *p;
	int e;

	/*
	 * Setting up timeouts.
	 */
	p = getenv("URUN_TIMEOUT");
	if(p) {
		urun_timeout = atoi(p);
		if(urun_timeout < 1)
			urun_timeout = 1;
	} else {
		urun_timeout = 3600;
	}
	max_allowed_time = time(NULL) + urun_timeout;

	tmo.tv_sec = normal_timeout;
	tmo.tv_nsec = 0;


	for(;;) {

		/*
		 * Go into a war mode if limit is reached.
		 */
		if(war_mode != 1 && time(NULL) > max_allowed_time) {
			if(war_mode == 0) {
				char msg[] = "\n Session limit is reached\n";
				write(2, msg, sizeof(msg) - 1);
			}

			war_mode = 1;

			/*
			 * Go over stored pids killing them.
			 */
			kill_recorded_processes(1 /* Allow warning */);
		}

		/*
		 * Gather events.
		 */
		ret = kevent(q,
			NULL, 0,
			kevl, sizeof(kevl)/sizeof(kevl[0]),
			&tmo);
		if(ret <= 0) {
			if(ret == 0) {
				if(running_pids_count == 0)
					/* There are no processes anymore! */
					_exit(0);

				if(war_mode) {
					/*
					 * Very funny: got a timeout
					 * in a war mode. Let's reintroduce
					 * the war mode.
					 */
					war_mode++;
				}
			} else if(ret == -1) {
				if(errno == EINTR)
					continue;
				break;
			}

		}

		if(running_pids_count == 0) {
			/*
			 * We were out of processes, but got someting.
			 * Probably this is a delayed notification.
			 * Return the normal timeout back.
			 */
			tmo.tv_sec = normal_timeout;
		}

		/*
		 * Process event.
		 */
		for(e = 0; e < ret; e++) {
			pid_t pid = (pid_t)kevl[e].ident;
			int flags = kevl[e].flags;

			if(flags & NOTE_TRACK && kevl[e].fflags & NOTE_CHILD) {
				/* Process has come */
				proc_status(pid, 1);
				if(war_mode) {
					/*
					 * Something has JUST started.
					 * It might be safe to use SIGKILL
					 * on that. Anyway, play hard.
					 */
					kill(pid, SIGKILL);
					/*
					 * Reintroduce war mode.
					 */
					war_mode++;
				}
			}

			if(flags & EV_EOF) {
				/* Process has died */
				proc_status(pid, 0);

				if(!running_pids_count) {
					/* Wait a little bit */
					tmo.tv_sec = 1;
				}
			}

			if(flags & NOTE_TRACKERR) {
				/*
				 * Unable to attach an event
				 * to the child process.
				 * Go into war mode.
				 */
				war_mode++;
			}
		}

	}

	/*
	 * We could end up here if kevent() is not working.
	 * Say the last word.
	 */
	if(running_pids_count) {
		kill_recorded_processes(1);
		sleep(5);
		kill_recorded_processes(0);
	}
}

/*
 * Record the process state.
 */
int
proc_status(pid_t pid, int status) {
	int bucket;
	pidstore_t *pst;

	if(pid <= 1) {
		/* Don't give a shit. */
		errno = ERANGE;
		return -1;
	}

	bucket = pid % RP_BUCKETS;
	if(status) {
		/*
		 * Attach this pid to the list of runnable.
		 * Rely on the kqueue not returning the same info twice.
		 */
		if(free_psts) {
			/* Fetch from freed storage */
			pst = free_psts;
			free_psts = free_psts->next;
		} else {
			pst = malloc(sizeof *pst);
			if(!pst) {
				/* Sorry, no can do. */
				return -1;
			}
		}

		pst->pid = pid;
		pst->warned = 0;
		pst->next = running_pids[bucket].next;
		running_pids[bucket].next = pst;
		running_pids_count++;
	} else {
		pidstore_t *prev = &running_pids[bucket];
		int found = 0;

		for(pst = running_pids[bucket].next; pst;
				prev = pst, pst = pst->next) {

			if(pst->pid == pid) {
				/* Remove from the running threads storage */
				prev->next = pst->next;

				/* Attach to the list of freed */
				pst->next = free_psts;
				free_psts = pst;

				--running_pids_count;
				++found;

				/*
				 * We could break here,
				 * but let's waste 50% of performance
				 * searching for unlikely doubles.
				 */
			}

		}
		if(!found) {
			errno = ESRCH;
			return -1;
		}
	}

	return 0;
}


/*
 * Warn or kill the processes.
 */
void
kill_recorded_processes(int allow_warn) {
	int buck;

	for(buck = 0; buck < RP_BUCKETS; buck++) {
		pidstore_t *pst, *next;
		int sig;
		int ret;

		for(pst  = running_pids[buck].next; pst; pst = next) {
			next = pst->next;

			sig = (pst->warned++ || !allow_warn)
				? SIGKILL
				: SIGTERM;
			ret = kill(pst->pid, sig);

			if(ret == -1 && errno == ESRCH) {
				/*
				 * Missed the signal?
				 */
				if(next && next->pid != pst->pid) {
					proc_status(pst->pid, 0);
				} else {
					/* We could've trapped... */
				}
			}
		}

	}

}


